from collections.abc import AsyncIterator
import types
from functools import partial
from contextlib import asynccontextmanager, ExitStack

from asyncgui import _current_task, _sleep_forever, move_on_when, ExclusiveEvent, _wait_args


@types.coroutine
def event(event_dispatcher, event_name, *, filter=None, stop_dispatching=False):
    '''
    Returns an :class:`~collections.abc.Awaitable` that can be used to wait for:

    * a Kivy event to occur.
    * a Kivy property's value to change.

    .. code-block::

        # Wait for a button to be pressed.
        await event(button, 'on_press')

        # Wait for an 'on_touch_down' event to occur.
        __, touch = await event(widget, 'on_touch_down')

        # Wait for 'widget.x' to change.
        __, x = await ak.event(widget, 'x')

    The ``filter`` parameter:

    .. code-block::

        # Wait for an 'on_touch_down' event to occur inside a widget.
        __, touch = await event(widget, 'on_touch_down', filter=lambda w, t: w.collide_point(*t.opos))

        # Wait for 'widget.x' to become greater than 100.
        if widget.x <= 100:
            await event(widget, 'x', filter=lambda __, x: x > 100)

    The ``stop_dispatching`` parameter:

      This only works for events not for properties.
    '''
    task = (yield _current_task)[0][0]
    bind_id = event_dispatcher.fbind(event_name, partial(_event_callback, filter, task._step, stop_dispatching))
    assert bind_id  # check if binding succeeded
    try:
        return (yield _sleep_forever)[0]
    finally:
        event_dispatcher.unbind_uid(event_name, bind_id)


def _event_callback(filter, task_step, stop_dispatching, *args, **kwargs):
    if (filter is None) or filter(*args, **kwargs):
        task_step(*args)
        return stop_dispatching


class event_freq:
    '''
    When handling a frequently occurring event, such as ``on_touch_move``, the following kind of code *might* cause
    performance issues:

    .. code-block::

        __, touch = await event(widget, 'on_touch_down')

        # This loop registers and unregisters an event handler on every iteration.
        while True:
            await event(widget, 'on_touch_move', filter=lambda w, t: t is touch)
            ...

    If that happens, try the following code instead. It might resolve the issue:

    .. code-block::

        __, touch = await event(widget, 'on_touch_down')
        async with event_freq(widget, 'on_touch_move', filter=lambda w, t: t is touch) as on_touch_move:
            while True:
                await on_touch_move()
                ...

    .. versionadded:: 0.7.1

    .. versionchanged:: 0.9.0
        The ``free_to_await`` parameter was added.

    The ``free_to_await`` parameter:

    If set to False (the default), the only permitted async operation within the with-block is ``await xxx()``,
    where ``xxx`` is the identifier specified in the as-clause. To lift this restriction, set ``free_to_await`` to
    True — at the cost of slightly reduced performance.
    '''
    __slots__ = ('_disp', '_name', '_filter', '_stop', '_bind_id', '_free_to_await')

    def __init__(self, event_dispatcher, event_name, *, filter=None, stop_dispatching=False, free_to_await=False):
        self._disp = event_dispatcher
        self._name = event_name
        self._filter = filter
        self._stop = stop_dispatching
        self._free_to_await = free_to_await

    @types.coroutine
    def __aenter__(self):
        if self._free_to_await:
            e = ExclusiveEvent()
            self._bind_id = self._disp.fbind(self._name, partial(_event_callback, self._filter, e.fire, self._stop))
            return e.wait_args
        else:
            task = (yield _current_task)[0][0]
            self._bind_id = self._disp.fbind(
                self._name, partial(_event_callback, self._filter, task._step, self._stop))
            return _wait_args

    async def __aexit__(self, *args):
        self._disp.unbind_uid(self._name, self._bind_id)


class suppress_event:
    '''
    Returns a context manager that prevents the callback functions (including the default handler) bound to an event
    from being called.

    .. code-block::
        :emphasize-lines: 4

        from kivy.uix.button import Button

        btn = Button()
        btn.bind(on_press=lambda __: print("pressed"))
        with suppress_event(btn, 'on_press'):
            btn.dispatch('on_press')

    The above code prints nothing because the callback function won't be called.

    Strictly speaking, this context manager doesn't prevent all callback functions from being called.
    It only prevents the callback functions that were bound to an event before the context manager enters.
    Thus, the following code prints ``pressed``.

    .. code-block::
        :emphasize-lines: 5

        from kivy.uix.button import Button

        btn = Button()
        with suppress_event(btn, 'on_press'):
            btn.bind(on_press=lambda __: print("pressed"))
            btn.dispatch('on_press')
    '''
    __slots__ = ('_dispatcher', '_name', '_bind_uid', '_filter', )

    def __init__(self, event_dispatcher, event_name, *, filter=lambda *args, **kwargs: True):
        self._dispatcher = event_dispatcher
        self._name = event_name
        self._filter = filter

    def __enter__(self):
        self._bind_uid = self._dispatcher.fbind(self._name, self._filter)

    def __exit__(self, *args):
        self._dispatcher.unbind_uid(self._name, self._bind_uid)


class block_touch_events:
    '''
    .. code-block::

        with block_touch_events(widget):
            ...

    Returns a context manager that blocks all touch events that meet **both** of the following criteria:

    * The touch is not currently grabbed by any widget. (i.e. ``touch.grab_current is None``)
    * The touch is inside the widget's bounding box. (i.e. ``widget.collide_point(*touch.pos)``)

    Basically equivalent to the following:

    .. code-block::

        def f(w, t):
            return t.grab_current is None and w.collide_point(*t.pos)
        with (
            suppress_event(widget, 'on_motion', filter=f),
            suppress_event(widget, 'on_touch_down', filter=f),
            suppress_event(widget, 'on_touch_move', filter=f),
            suppress_event(widget, 'on_touch_up', filter=f),
        ):
            ...

    .. versionadded:: 0.10.0
    '''
    __slots__ = ('_dispatcher', '_filter', )

    def __init__(self, event_dispatcher, *, filter=lambda w, t: t.grab_current is None and w.collide_point(*t.pos)):
        self._dispatcher = event_dispatcher
        self._filter = filter

    def __enter__(self):
        f = self._filter
        self._dispatcher.bind(on_motion=f, on_touch_down=f, on_touch_move=f, on_touch_up=f)

    def __exit__(self, *__):
        f = self._filter
        self._dispatcher.unbind(on_motion=f, on_touch_down=f, on_touch_move=f, on_touch_up=f)


async def rest_of_touch_events(widget, touch, *, stop_dispatching=False, grab=True) -> AsyncIterator[None]:
    '''
    Returns an async iterator that yields None on each ``on_touch_move`` event
    and stops when an ``on_touch_up`` event occurs.

    .. code-block::

        async for __ in rest_of_touch_events(widget, touch):
            print('on_touch_move')
        print('on_touch_up')

    :param grab: If set to ``False``, this API will not rely on ``touch.grab()``, which means there is no guarantee
                 that all events from the given touch will be delivered to the widget, `as documented <grab_>`_.
                 If the ``on_touch_up`` event is not delivered, the iterator will wait indefinitely for it—an event
                 that never comes. Do not set this to ``False`` unless you know what you are doing.
    :param stop_dispatching: Whether to stop dispatching non-grabbed touch events.
                             (Grabbed events are always stopped if the ``grab`` is ``True``.)
                             For details, see `event-bubbling`_.

    .. warning::
        You should not use this when Kivy is running in async mode. Use :func:`rest_of_touch_events_cm` instead.

    .. versionchanged:: 0.9.0
        The ``timeout`` parameter was removed.

    .. versionchanged:: 0.9.1
        The ``grab`` parameter was added.

    .. _grab: https://kivy.org/doc/master/guide/inputs.html#grabbing-touch-events
    .. _event-bubbling: https://kivy.org/doc/master/api-kivy.uix.widget.html#widget-touch-event-bubbling
    '''
    async with rest_of_touch_events_cm(widget, touch, stop_dispatching=stop_dispatching, grab=grab) as on_touch_move:
        while True:
            await on_touch_move()
            yield


@asynccontextmanager
async def rest_of_touch_events_cm(widget, touch, *, stop_dispatching=False, free_to_await=False, grab=True):
    '''
    A variant of :func:`rest_of_touch_events`.
    This version is more verbose, but remains safe even when Kivy is running in async mode.

    .. code-block::

        async with rest_of_touch_events_cm(widget, touch) as on_touch_move:
            while True:
                await on_touch_move()
                print('on_touch_move')
        print('on_touch_up')

    .. versionadded:: 0.9.1
    '''
    def is_the_same_touch(w, t, touch=touch):
        return t is touch
    with ExitStack() as stack:
        if grab:
            touch.grab(widget)
            stack.callback(touch.ungrab, widget)
            if stop_dispatching:
                ec = stack.enter_context
                se = partial(suppress_event, widget, filter=is_the_same_touch)
                ec(se('on_touch_up'))
                ec(se('on_touch_move'))

            def filter(w, t, touch=touch):
                return t is touch and t.grab_current is w
            stop_dispatching = True
        else:
            filter = is_the_same_touch
        async with (
            move_on_when(event(widget, 'on_touch_up', filter=filter, stop_dispatching=stop_dispatching)),
            event_freq(widget, 'on_touch_move', filter=filter, stop_dispatching=stop_dispatching,
                       free_to_await=free_to_await) as on_touch_move,
        ):
            yield on_touch_move



__all__ = (
    'anim_attrs',
    'anim_attrs_abbr',
    'anim_with_ratio',
    'block_touch_events',
    'event',
    'event_freq',
    'fade_transition',
    'interpolate',
    'interpolate_seq',
    'managed_start',
    'move_on_after',
    'n_frames',
    'rest_of_touch_events',
    'rest_of_touch_events_cm',
    'run_in_executor',
    'run_in_thread',
    'sandwich_canvas',
    'sleep',
    'sleep_free',
    'sleep_freq',
    'smooth_attr',
    'stencil_mask',
    'stencil_widget_mask',
    'suppress_event',
    'sync_attr',
    'sync_attrs',
    'transform',
)

from asyncgui import *
from ._sleep import sleep, sleep_free, move_on_after, n_frames, sleep_freq, anim_with_ratio
from ._event import event, event_freq, suppress_event, rest_of_touch_events, rest_of_touch_events_cm, \
    block_touch_events
from ._anim_attrs import anim_attrs, anim_attrs_abbr
from ._interpolate import interpolate, interpolate_seq, fade_transition
from ._threading import run_in_executor, run_in_thread
from ._etc import transform, sync_attr, sync_attrs, stencil_mask, stencil_widget_mask, sandwich_canvas, smooth_attr
from ._managed_start import managed_start



from contextlib import AbstractAsyncContextManager
import types

from kivy.clock import Clock
from asyncgui import _current_task, _sleep_forever, move_on_when, Task, Cancelled, ExclusiveEvent, _wait_args_0


@types.coroutine
def sleep(duration):
    '''
    An async form of :meth:`kivy.clock.Clock.schedule_once`.

    .. code-block::

        dt = await sleep(5)  # wait for 5 seconds
    '''
    task = (yield _current_task)[0][0]
    clock_event = Clock.create_trigger(task._step, duration, False, False)
    clock_event()

    try:
        return (yield _sleep_forever)[0][0]
    except Cancelled:
        clock_event.cancel()
        raise


@types.coroutine
def sleep_free(duration):
    '''
    An async form of :meth:`kivy.clock.Clock.schedule_once_free`.

    .. code-block::

        dt = await sleep_free(5)  # wait for 5 seconds
    '''
    task = (yield _current_task)[0][0]
    clock_event = Clock.create_trigger_free(task._step, duration, False, False)
    clock_event()

    try:
        return (yield _sleep_forever)[0][0]
    except Cancelled:
        clock_event.cancel()
        raise


class sleep_freq:
    '''
    An async form of :meth:`kivy.clock.Clock.schedule_interval`. The following callback-style code:

    .. code-block::

        def callback(dt):
            print(dt)
            if some_condition:
                return False

        Clock.schedule_interval(callback, 0.1)

    is equivalent to the following async-style code:

    .. code-block::

        async with sleep_freq(0.1) as sleep:
            while True:
                dt = await sleep()
                print(dt)
                if some_condition:
                    break

    .. versionchanged:: 0.8.0
        The API was made private.

    .. versionchanged:: 0.9.0

        * The API was made public again.
        * The ``free_to_await`` parameter was added.

    The ``free_to_await`` parameter:

    If set to False (the default), the only permitted async operation within the with-block is ``await xxx()``,
    where ``xxx`` is the identifier specified in the as-clause. To lift this restriction, set ``free_to_await`` to
    True — at the cost of slightly reduced performance.
    '''

    __slots__ = ('_step', '_trigger', '_free_to_await')

    def __init__(self, step=0, free_to_await=False):
        self._step = step
        self._free_to_await = free_to_await

    @types.coroutine
    def __aenter__(self):
        if self._free_to_await:
            e = ExclusiveEvent()
            self._trigger = t = Clock.create_trigger(e.fire, self._step, True, False)
            t()
            return e.wait_args_0
        else:
            task = (yield _current_task)[0][0]
            self._trigger = t = Clock.create_trigger(task._step, self._step, True, False)
            t()
            return _wait_args_0

    async def __aexit__(self, *args):
        self._trigger.cancel()


async def anim_with_ratio(*, base, step=0):
    '''
    Returns an async iterator that yields the elapsed time since the start of the iteration, divided by ``base``.

    .. code-block::

        async for p in anim_with_ratio(base=3):
            print(p)

    The code above is equivalent to the following:

    .. code-block::

        async with sleep_freq() as sleep:
            base = 3
            total_elapsed_time = 0.
            while True:
                total_elapsed_time += await sleep()
                p = total_elapsed_time / base
                print(p)

    Use :class:`kivy.animation.AnimationTransition` for non-linear curves.

    .. code-block::

        from kivy.animation import AnimationTransition

        in_cubic = AnimationTransition.in_cubic

        async for p in anim_with_ratio(base=...):
            p = in_cubic(p)
            print(p)

    .. versionadded:: 0.6.1

    .. versionchanged:: 0.7.0

        The ``duration`` parameter was replaced with ``base``.
        The loop no longer ends on its own.
    '''
    async with sleep_freq(step=step) as sleep:
        et = 0.
        while True:
            et += await sleep()
            yield et / base


def move_on_after(seconds: float) -> AbstractAsyncContextManager[Task]:
    '''
    Returns an async context manager that applies a time limit to its code block,
    like :func:`trio.move_on_after` does.

    .. code-block::

        async with move_on_after(seconds) as timeout_tracker:
            ...
        if timeout_tracker.finished:
            print("The code block was interrupted due to a timeout")
        else:
            print("The code block exited gracefully.")

    .. versionadded:: 0.6.1
    '''
    return move_on_when(sleep(seconds))


@types.coroutine
def n_frames(n: int):
    '''
    Waits for a specified number of frames to elapse.

    .. code-block::

        await n_frames(2)

    If you want to wait for one frame, :func:`asynckivy.sleep` is preferable for a performance reason.

    .. code-block::

        await sleep(0)
    '''
    if n < 0:
        raise ValueError(f"Waiting for {n} frames doesn't make sense.")
    if not n:
        return

    task = (yield _current_task)[0][0]

    def callback(dt):
        nonlocal n
        n -= 1
        if not n:
            task._step()
            return False

    clock_event = Clock.schedule_interval(callback, 0)

    try:
        yield _sleep_forever
    finally:
        clock_event.cancel()



from threading import Thread
from concurrent.futures import ThreadPoolExecutor
from kivy.clock import Clock
import asyncgui


def _wrapper(func, ev):
    ret = None
    exc = None
    try:
        ret = func()
    except Exception as e:
        exc = e
    finally:
        Clock.schedule_once(lambda __: ev.fire(ret, exc))


async def run_in_thread(func, *, daemon=None):
    '''
    Creates a new thread, runs a function within it, then waits for the completion of that function.

    .. code-block::

        return_value = await run_in_thread(func)

    See :ref:`io-in-asynckivy` for details.

    .. warning::
        When the caller Task is cancelled, the ``func`` will be left running, which violates "structured concurrency".
    '''
    ev = asyncgui.ExclusiveEvent()
    Thread(
        name='asynckivy.run_in_thread',
        target=_wrapper, daemon=daemon, args=(func, ev, ),
    ).start()
    ret, exc = (await ev.wait())[0]
    if exc is not None:
        raise exc
    return ret


async def run_in_executor(executor: ThreadPoolExecutor, func):
    '''
    Runs a function within a :class:`concurrent.futures.ThreadPoolExecutor`, and waits for the completion of the
    function.

    .. code-block::

        executor = ThreadPoolExecutor()
        ...
        return_value = await run_in_executor(executor, func)

    See :ref:`io-in-asynckivy` for details.

    .. warning::
        When the caller Task is cancelled, the ``func`` will be left running if it has already started,
        which violates "structured concurrency".
    '''
    ev = asyncgui.ExclusiveEvent()
    future = executor.submit(_wrapper, func, ev)
    try:
        ret, exc = (await ev.wait())[0]
    except asyncgui.Cancelled:
        future.cancel()
        raise
    assert future.done()
    if exc is not None:
        raise exc
    return ret



import types
from functools import partial
import kivy.clock
from kivy.animation import AnimationTransition
import asyncgui


def _update(setattr, zip, min, obj, duration, transition, anim_params, task, p_time, dt):
    time = p_time[0] + dt
    p_time[0] = time

    # calculate progression
    progress = min(1., time / duration)
    t = transition(progress)

    # apply progression on obj
    for attr_name, org_value, slope, is_seq in anim_params:
        if is_seq:
            new_value = [
                slope_elem * t + org_elem
                for org_elem, slope_elem in zip(org_value, slope)
            ]
            setattr(obj, attr_name, new_value)
        else:
            setattr(obj, attr_name, slope * t + org_value)

    # time to stop ?
    if progress >= 1.:
        task._step()
        return False


_update = partial(_update, setattr, zip, min)


@types.coroutine
def _anim_attrs(
        obj, duration, step, transition, animated_properties,
        getattr=getattr, isinstance=isinstance, tuple=tuple, str=str, partial=partial, native_seq_types=(tuple, list),
        zip=zip, Clock=kivy.clock.Clock, AnimationTransition=AnimationTransition,
        _update=_update, _current_task=asyncgui._current_task, _sleep_forever=asyncgui._sleep_forever, /):
    if isinstance(transition, str):
        transition = getattr(AnimationTransition, transition)

    # get current values & calculate slopes
    anim_params = [
        (
            org_value := getattr(obj, attr_name),
            is_seq := isinstance(org_value, native_seq_types),
            (
                org_value := tuple(org_value),
                slope := [goal_elem - org_elem for goal_elem, org_elem in zip(goal_value, org_value)],
            ) if is_seq else (slope := goal_value - org_value),
        ) and (attr_name, org_value, slope, is_seq, )
        for attr_name, goal_value in animated_properties.items()
    ]

    try:
        clock_event = Clock.schedule_interval(
            partial(_update, obj, duration, transition, anim_params, (yield _current_task)[0][0], [0., ]),
            step,
        )
        yield _sleep_forever
    finally:
        clock_event.cancel()


def anim_attrs(obj, *, duration=1.0, step=0, transition=AnimationTransition.linear, **animated_properties):
    '''
    Animates attibutes of any object.

    .. code-block::

        import types

        obj = types.SimpleNamespace(x=0, size=(200, 300))
        await anim_attrs(obj, x=100, size=(400, 400))

    .. warning::

        Unlike :class:`kivy.animation.Animation`, this one does not support dictionary-type and nested-sequence.

        .. code-block::

            await anim_attrs(obj, pos_hint={'x': 1.})  # not supported
            await anim_attrs(obj, nested_sequence=[[10, 20, ]])  # not supported

            await anim_attrs(obj, color=(1, 0, 0, 1), pos=(100, 200))  # OK

    .. versionadded:: 0.6.1
    .. versionchanged:: 0.9.0
        The ``output_seq_type`` parameter was removed.
    '''
    return _anim_attrs(obj, duration, step, transition, animated_properties)


def anim_attrs_abbr(obj, *, d=1.0, s=0, t=AnimationTransition.linear, **animated_properties):
    '''
    :func:`anim_attrs` cannot animate attributes named ``step``, ``duration`` and ``transition`` but this one can.

    .. versionadded:: 0.6.1
    .. versionchanged:: 0.9.0
        The ``output_seq_type`` parameter was removed.
    '''
    return _anim_attrs(obj, d, s, t, animated_properties)



__all__ = (
    'fade', 'fade_multiple', 'slide', 'scale', 'iris', 'shader', 'gl_transitions_dot_com',
)

from typing import Literal, TypeAlias, Union
from collections.abc import Sequence, Mapping
from contextlib import asynccontextmanager, ExitStack
import math
import string

from kivy.utils import colormap
from kivy.graphics.texture import Texture
from kivy.graphics import (
    Translate, Scale, InstructionGroup, StencilPop, StencilPush, StencilUnUse, StencilUse,
    Rectangle, Ellipse, Color, VertexInstruction, Canvas, BindTexture, Fbo, RenderContext,
)
from kivy.animation import AnimationTransition
from kivy.core.window import Window, WindowBase
from kivy.uix.widget import Widget

import asynckivy as ak
from asynckivy import transform, anim_attrs_abbr as anim_attrs, sleep_freq


linear = AnimationTransition.linear
Wow: TypeAlias = Union[WindowBase, Widget]  # 'Wow' -> WindowBase or Widget


@asynccontextmanager
async def fade(target: Wow | Canvas=Window.canvas, *, goal_opacity=0., duration=1., out_curve=linear, in_curve=linear):
    '''
    Fades out the ``target``, executes the code inside the with-block, and then fades it back in.

    .. versionadded:: 0.9.0
    .. versionchanged:: 0.9.1
        Now accepts Canvas as target.
    '''
    half_d = duration / 2
    original = target.opacity
    try:
        await anim_attrs(target, d=half_d, t=out_curve, opacity=goal_opacity)
        yield
        await anim_attrs(target, d=half_d, t=in_curve, opacity=original)
    finally:
        target.opacity = original


@asynccontextmanager
async def fade_multiple(*widgets, duration=1., out_curve=linear, in_curve=linear):
    '''
    Fades out the ``widgets`` simultaneously, executes the code inside the with-block, and then fades them back in.

    .. versionadded:: 0.10.0
    '''
    zip_ = zip
    half_d = duration / 2.
    orig_opacities = [w.opacity for w in widgets]
    if isinstance(out_curve, str):
        out_curve = getattr(AnimationTransition, out_curve)
    if isinstance(in_curve, str):
        in_curve = getattr(AnimationTransition, in_curve)
    try:
        async with sleep_freq() as sleep:
            elapsed_time = 0.
            while True:
                elapsed_time += await sleep()
                p = 1.0 - out_curve(elapsed_time / half_d)
                for w, o in zip_(widgets, orig_opacities):
                    w.opacity = p * o
                if elapsed_time >= half_d:
                    break
        yield
        async with sleep_freq() as sleep:
            elapsed_time = 0.
            while elapsed_time < half_d:
                elapsed_time += await sleep()
                p = in_curve(elapsed_time / half_d)
                for w, o in zip_(widgets, orig_opacities):
                    w.opacity = p * o
    finally:
        for w, o in zip_(widgets, orig_opacities):
            w.opacity = o


@asynccontextmanager
async def slide(target: Wow=Window, *, duration=1., out_curve='in_back', in_curve='out_back',
                x_direction: Literal['left', 'right', None]='left',
                y_direction: Literal['down', 'up', None]=None,
                working_layer="inner"):
    '''
    Slides the ``target`` out, executes the code inside the with-block, and then slides it back in.

    .. versionadded:: 0.9.0
    .. versionchanged:: 0.9.2
        The default value of ``out_curve`` has changed from ``'in_cubic'`` to ``'in_back'``.
        The default value of ``in_curve`` has changed from ``'out_cubic'`` to ``'out_back'``.
    '''
    x_dist, y_dist = target.size
    if x_direction is None:
        x_dist = 0
    elif x_direction == 'left':
        x_dist = -x_dist
    if y_direction is None:
        y_dist = 0
    elif y_direction == 'down':
        y_dist = -y_dist

    with transform(target, working_layer=working_layer) as ig:
        ig.add(mat := Translate())
        half_d = duration / 2
        await anim_attrs(mat, d=half_d, t=out_curve, x=x_dist, y=y_dist)
        yield
        mat.x = -x_dist
        mat.y = -y_dist
        await anim_attrs(mat, d=half_d, t=in_curve, x=0, y=0)


@asynccontextmanager
async def scale(target: Wow=Window, *, duration=1, out_curve='out_quad', in_curve='in_quad',
                working_layer="inner"):
    '''
    Shrinks the ``target``, executes the code inside the with-block, and then restores it to its original size.

    .. versionadded:: 0.9.0
    '''
    with transform(target, working_layer=working_layer) as ig:
        ig.add(mat := Scale(origin=target.center))
        half_d = duration / 2
        await anim_attrs(mat, d=half_d, t=out_curve, xyz=(0, 0, 1))
        yield
        await anim_attrs(mat, d=half_d, t=in_curve, xyz=(1, 1, 1))


def _calc_enclosing_circle_radius(circle_center, rectangle_size, max=max, hypot=math.hypot, abs=abs):
    '''
    Calculates the minimum radius required for a circle centered at the given position to fully
    enclose a rectangle of the given size, assuming its bottom-left corner is at (0, 0).

    .. code-block::

        radius = _calc_enclosing_circle_radius(circle_center, rectangle_size)
    '''
    w, h = rectangle_size
    x, y = circle_center
    return hypot(max(abs(x), abs(w - x)), max(abs(y), abs(h - y)))


@asynccontextmanager
async def iris(target: WindowBase=Window, *, duration=1, out_curve='in_cubic', in_curve='out_cubic',
               color: Sequence[float]=colormap['white'], circle_center: Sequence[float]=None,
               overlay: VertexInstruction=None):
    '''
    縮む円によって ``target`` が見える範囲を絞っていき、完全に見えなくなったらwithブロック内を実行し、その後円を広げて ``target`` を再び見せる。

    .. versionadded:: 0.9.0
    '''
    if not isinstance(target, WindowBase):
        raise TypeError(f"'target' must be a WindowBase instance, not {type(target).__name__}")
    half_d = duration / 2
    canvas = target.canvas
    if circle_center is None:
        circle_center = target.center
    if overlay is None:
        overlay = Rectangle(size=target.size)
    radius = _calc_enclosing_circle_radius(circle_center, target.size)
    diameter = radius * 2.
    ellipse_start_pos = (circle_center[0] - radius, circle_center[1] - radius)
    ellipse_start_size = (diameter, diameter)
    ig = InstructionGroup()
    ig_add = ig.add
    ig_add(StencilPush())
    ig_add(inner_ig := InstructionGroup())
    inner_ig.add(ellipse := Ellipse(pos=ellipse_start_pos, size=ellipse_start_size))
    ig_add(overlay)
    ig_add(StencilUse())
    ig_add(Color(*color))
    ig_add(overlay)
    ig_add(StencilUnUse())
    ig_add(overlay)
    ig_add(inner_ig)
    ig_add(StencilPop())

    canvas.add(ig)
    try:
        await anim_attrs(ellipse, d=half_d, t=out_curve, pos=circle_center, size=(0, 0))
        # Setting the ellipse size to (0, 0) isn't enough to nullify the stencil effect for some reason,
        # so we remove it from the canvas and re-add it later.
        inner_ig.remove(ellipse)
        yield
        inner_ig.add(ellipse)
        await anim_attrs(ellipse, d=half_d, t=in_curve, pos=ellipse_start_pos, size=ellipse_start_size)
    finally:
        canvas.remove(ig)


def _render_to_texture(widget) -> Texture:
    fbo = Fbo(size=widget.size)
    fbo.children.extend(
        (Translate(-widget.x, -widget.y, 0.), widget.canvas, )
    )
    fbo.draw()
    fbo.children.clear()
    return fbo.texture


@asynccontextmanager
async def shader(target: Widget, fs: str, *, duration=1., uniforms: Mapping=None,
                 progress_var: str='progress', out_texture_var: str='out_tex', in_texture_var: str='in_tex'):
    '''
    A transition that uses a GLSL fragment shader.

    .. code-block::

        fade = """
        /* Outputs from the vertex shader */
        varying vec4 frag_color;
        varying vec2 tex_coord0;

        uniform float t;
        uniform sampler2D out_tex;
        uniform sampler2D in_tex;

        void main(void) {
            vec4 cout = texture2D(out_tex, tex_coord0);
            vec4 cin = texture2D(in_tex, tex_coord0);
            gl_FragColor = mix(cout, cin, t);
        }
        """

        async with shader(widget, fs=fade, progress_var='t', out_texture_var='out_tex', in_texture_var='in_tex'):
            ...

    :param fs: The GLSL fragment shader code to use.
    :param progress_var: Name of the uniform ``float`` variable representing the normalized progress
                         (elapsed time divided by ``duration``).
    :param out_texture_var: Name of the uniform ``sampler2D`` variable for the *outgoing* texture.
                            This corresponds to the rendered content of the target before ``__aenter__``.
    :param in_texture_var: Name of the uniform ``sampler2D`` variable for the *incoming* texture.
                           This corresponds to the rendered content of the target before ``__aexit__``.
    :param uniforms: Additional uniform variables to be set on the shader.

    .. warning::
        Do not modify the target's parent — such as adding or removing widgets or graphics instructions — while the
        transition is in progress. Modifying the target itself is fine.
    .. versionadded:: 0.9.3
    '''
    parent_canvas = target.parent.canvas
    original_idx = parent_canvas.children.index(target.canvas)

    with ExitStack() as stack:
        parent_canvas.remove(target.canvas)
        stack.callback(parent_canvas.insert, original_idx, target.canvas)
        out_texture = _render_to_texture(target)
        ig = InstructionGroup()
        parent_canvas.insert(original_idx, ig)
        try:
            ig.add(Color())
            ig.add(rect := Rectangle(size=target.size, pos=target.pos, texture=out_texture))
            yield
            await ak.sleep(0)
        finally:
            parent_canvas.remove(ig)
            ig.clear()

        in_texture = _render_to_texture(target)
        rc = RenderContext(fs=fs, use_parent_projection=True, use_parent_modelview=True)
        if not rc.shader.success:
            raise ValueError('Failed to set shader')
        parent_canvas.insert(original_idx, rc)
        stack.callback(parent_canvas.remove, rc)
        rc.add(BindTexture(texture=out_texture, index=1))
        rc.add(BindTexture(texture=in_texture, index=2))
        rect.texture = None
        rc.add(rect)
        rc[out_texture_var] = 1
        rc[in_texture_var] = 2
        rc[progress_var] = elapsed_time = 0.
        if uniforms:
            for k, v in uniforms.items():
                rc[k] = v

        async with ak.sleep_freq() as sleep:
            while elapsed_time < duration:
                elapsed_time += await sleep()
                rc[progress_var] = elapsed_time / duration


GL_TRANSITIONS_DOT_COM_TEMPLATE = string.Template('''
#ifdef GL_ES
    precision highp float;
#endif

/* Outputs from the vertex shader */
varying vec4 frag_color;
varying vec2 tex_coord0;

uniform float progress;
uniform sampler2D out_tex;
uniform sampler2D in_tex;

vec4 getFromColor(vec2 uv) {
    return texture2D(out_tex, uv);
}
vec4 getToColor(vec2 uv) {
    return texture2D(in_tex, uv);
}

${usercode}

void main(void) {
    gl_FragColor = transition(tex_coord0);
}
''')


def gl_transitions_dot_com(target: Widget, fs: str, *, duration=1., uniforms: Mapping=None):
    '''
    A transition that uses a GLSL fragment shader from https://gl-transitions.com/.

    .. code-block::

        cross_warp = """
        // Author: Eke Péter <peterekepeter@gmail.com>
        // License: MIT
        vec4 transition(vec2 p) {
            float x = progress;
            x=smoothstep(.0,1.0,(x*2.0+p.x-1.0));
            return mix(getFromColor((p-.5)*(1.-x)+.5), getToColor((p-.5)*x+.5), x);
        }
        """

        async with gl_transitions_dot_com(widget, fs=cross_warp):
            ...

    See :func:`shader` for details.

    .. versionadded:: 0.9.3
    '''
    return shader(
        target,
        GL_TRANSITIONS_DOT_COM_TEMPLATE.substitute(usercode=fs),
        progress_var='progress',
        out_texture_var='out_tex',
        in_texture_var='in_tex',
        duration=duration,
        uniforms=uniforms,
    )


# Aliases for backward compatibility
fade_transition = fade
slide_transition = slide
scale_transition = scale
iris_transition = iris
